/*
  SimpleUC1701.cpp
*/

#include <Arduino.h>

#include <SPI.h>

#include <SimpleUC1701.h>

uint8_t buf[ScreenWidth * ScreenHeight / 8];
int dontSendBuf;
bool MakeMark;
int Cursorx, Cursory;
uint8_t xOffset = 4;
uint8_t pinCS; // default 10
uint8_t pinCD; // default  7
uint8_t pinRST; // default  6
bool curxFlip;
bool curyFlip;

void UC1701SendCommand(uint8_t b) {
  digitalWrite(pinCD, 0);
  SPI.transfer(b);
}

void UC1701SendData(uint8_t b) {
  digitalWrite(pinCD, 1);
  SPI.transfer(b);
}

void SendBufxy(uint8_t x0, uint8_t y0, uint8_t x1, uint8_t y1) {
  int p, p0, p1, i, j, b, x;

  if (dontSendBuf > 0)
    exit;

  p0 = max(y0 / 8, 0);
  p1 = min(y1, ScreenVMax) / 8;

  for (p = p0; p <= p1; p++) {
    UC1701SendCommand(0x10 + (x0 + xOffset) / 16); // Set Column Address MSB
    UC1701SendCommand(0x00 + (x0 + xOffset) % 16); // Set Column Address LSB
    UC1701SendCommand(0xB0 + p); // Set Page Address

    for (x = x0; x <= x1; x++)
      UC1701SendData(buf[x + p * ScreenWidth]);

    UC1701SendCommand(0xA4); //     Set All-Pixel-ON
    UC1701SendCommand(0xAF); //     Set Display Enable
  }
}

void SetPixA(uint8_t x, uint8_t y) {
  if ((x >= 0) and(x <= ScreenHMax) and(y >= 0) and(y <= ScreenVMax))
    if (MakeMark) {
      buf[x + (y / 8) * ScreenWidth] = buf[x + (y / 8) * ScreenWidth] | (1 << (y % 8));
    } else {
      buf[x + (y / 8) * ScreenWidth] = buf[x + (y / 8) * ScreenWidth] & (~(1 << (y % 8)));
    }
}

void DrawLine(uint8_t x1, uint8_t y1, uint8_t x2, uint8_t y2) {
  int tmp, x, y, dx, dy, err, ystep, swapxy;
  swapxy = 0;

  if (x1 > x2) dx = x1 - x2;
  else dx = x2 - x1;
  if (y1 > y2) dy = y1 - y2;
  else dy = y2 - y1;

  if (dy > dx) {
    swapxy = 1;
    tmp = dx;
    dx = dy;
    dy = tmp;
    tmp = x1;
    x1 = y1;
    y1 = tmp;
    tmp = x2;
    x2 = y2;
    y2 = tmp;
  }
  if ((x1 > x2)) {
    tmp = x1;
    x1 = x2;
    x2 = tmp;
    tmp = y1;
    y1 = y2;
    y2 = tmp;
  }
  err = dx >> 1;
  if ((y2 > y1)) ystep = 1;
  else ystep = -1;
  y = y1;
  for (x = x1; x <= x2; x++) {
    if ((swapxy == 0)) {
      SetPixA(x, y);
    } else {
      SetPixA(y, x);
    }
    err -= dy;
    if (err < 0) {
      y += ystep;
      err += dx;
    }
  }

  if (swapxy == 0)
    SendBufxy(min(x1, x2), min(y1, y2), max(x1, x2), max(y1, y2));
  else
    SendBufxy(min(y1, y2), min(x1, x2), max(y1, y2), max(x1, x2));
}

void DrawPixel(uint8_t x, uint8_t y) {
  SetPixA(x, y);
  SendBufxy(x, y, x, y);
}

void UC1701SetCursor(uint8_t x, uint8_t y) {
  Cursorx = x;
  Cursory = y;
}

void UC1701ClearDisplay() {
  dontSendBuf = 0;
  memset(buf, 0, sizeof(buf));
  SendBufxy(0, 0, ScreenHMax, ScreenVMax);
  UC1701SetCursor(20, 20);
}

void DrawHLine(uint8_t x, uint8_t y, uint8_t w) {
  DrawLine(x, y, x + w - 1, y);
}

void DrawVLine(uint8_t x, uint8_t y, uint8_t h) {
  DrawLine(x, y, x, y + h - 1);
}

void DrawBox(uint8_t x, uint8_t y, uint8_t w, uint8_t h) {
  dontSendBuf++;
  for (uint8_t i = x; i <= x + w - 1; i++)
    DrawVLine(i, y, h);
  dontSendBuf--;
  SendBufxy(x, y, x + w - 1, y + h - 1);
}

void DrawFrame(uint8_t x, uint8_t y, uint8_t w, uint8_t h) {
  dontSendBuf++;
  DrawHLine(x, y, w);
  DrawHLine(x, y + h - 1, w);
  DrawVLine(x, y, h);
  DrawVLine(x + w - 1, y, h);
  dontSendBuf--;
  SendBufxy(x, y, x + w - 1, y + h - 1);
}

void draw_ellipse_section(int x, int y, int x0, int y0, bool filled) {
  if (filled) {
    DrawVLine(x0 + x, y0 - y, y + 1);
    DrawVLine(x0 - x, y0 - y, y + 1);
    DrawVLine(x0 + x, y0, y + 1);
    DrawVLine(x0 - x, y0, y + 1);
  } else {
    SetPixA(x0 + x, y0 - y);
    SetPixA(x0 - x, y0 - y);
    SetPixA(x0 + x, y0 + y);
    SetPixA(x0 - x, y0 + y);
  }
}

void DrawEllipseEx(int x0, int y0, int rx, int ry, bool filled) {
  int x, y, xchg, ychg, err, rxrx2, ryry2, stopx, stopy;
  dontSendBuf++;

  rxrx2 = rx;
  rxrx2 = rxrx2 * rx;
  rxrx2 = rxrx2 * 2;
  ryry2 = ry;
  ryry2 = ryry2 * ry;
  ryry2 = ryry2 * 2;

  x = rx;
  y = 0;

  xchg = 1;
  xchg -= rx;
  xchg -= rx;
  xchg = xchg * ry;
  xchg = xchg * ry;

  ychg = rx;
  ychg = ychg * rx;

  err = 0;

  stopx = ryry2;
  stopx = stopx * rx;
  stopy = 0;

  while (stopx >= stopy) {
    draw_ellipse_section(x, y, x0, y0, filled);
    y++;
    stopy += rxrx2;
    err += ychg;
    ychg += rxrx2;
    if ((2 * err + xchg > 0)) {
      x--;
      stopx -= ryry2;
      err += xchg;
      xchg += ryry2;
    }
  }

  x = 0;
  y = ry;

  xchg = ry;
  xchg = xchg * ry;

  ychg = 1;
  ychg -= ry;
  ychg -= ry;
  ychg = ychg * rx;
  ychg = ychg * rx;

  err = 0;

  stopx = 0;

  stopy = rxrx2;
  stopy = stopy * ry;

  while (stopx <= stopy) {
    draw_ellipse_section(x, y, x0, y0, filled);
    x++;
    stopx += ryry2;
    err += xchg;
    xchg += ryry2;
    if ((2 * err + ychg > 0)) {
      y--;
      stopy -= rxrx2;
      err += ychg;
      ychg += rxrx2;
    }
  }

  dontSendBuf--;
  SendBufxy(x0 - rx, y0 - ry, x0 + rx, y0 + ry);
}

void DrawCircle(uint8_t x0, uint8_t y0, uint8_t rad) {
  DrawEllipse(x0, y0, rad, rad);
}

void DrawDisc(uint8_t x0, uint8_t y0, uint8_t rad) {
  DrawFilledEllipse(x0, y0, rad, rad);
}

void DrawFilledEllipse(int x0, int y0, int rx, int ry) {
  DrawEllipseEx(x0, y0, rx, ry, true);
}

void DrawEllipse(int x0, int y0, int rx, int ry) {
  DrawEllipseEx(x0, y0, rx, ry, false);
}

int CrossProduct(int leftx, int lefty, int rightx, int righty) {
  return (leftx * righty) - (lefty * rightx);
}

void DrawTriangle(uint8_t x0, uint8_t y0, uint8_t x1, uint8_t y1, uint8_t x2, uint8_t y2) {
  int minx, miny, maxx, maxy, s, t, u, x, y, qx, vs1x, vs2x, qy, vs1y, vs2y;
  dontSendBuf++;
  vs1x = x1 - x0;
  vs1y = y1 - y0;
  vs2x = x2 - x0;
  vs2y = y2 - y0;
  u = CrossProduct(vs1x, vs1y, vs2x, vs2y);

  minx = min(x0, min(x1, x2));
  maxx = max(x0, max(x1, x2));
  miny = min(y0, min(y1, y2));
  maxy = max(y0, max(y1, y2));
  for (x = minx; x <= maxx; x++) {
    for (y = miny; y <= maxy; y++) {
      qx = x - x0;
      qy = y - y0;

      s = CrossProduct(qx, qy, vs2x, vs2y);
      t = CrossProduct(vs1x, vs1y, qx, qy);

      if (u < 0) {
        if ((s <= 0) && (t <= 0) && ((s + t) >= u))
          SetPixA(x, y);
      } else {
        if ((s >= 0) && (t >= 0) && ((s + t) <= u))
          SetPixA(x, y);
      }
    }
  }
  DrawLine(x0,y0,x1,y1);
  DrawLine(x1,y1,x2,y2);
  DrawLine(x2,y2,x0,y0);
  dontSendBuf--;
  SendBufxy(minx, miny, maxx, maxy);
}

void DrawChar(uint8_t c, word Font) {
  word b, n, j, px;
  byte sz, desc;

  sz = pgm_read_byte_near(Font);
  Font++;
  desc = pgm_read_byte_near(Font);
  Font++;

  j = pgm_read_byte_near(Font); // fist char
  Font++;
  if (c < j) return;

  while (c > j) {
    b = pgm_read_byte_near(Font);
    if (b == 0)
      return;
    if (sz > 8)
      Font += b * 2 + 1;
    else
      Font += b + 1;
    c--;
  }
  px = Cursorx;
  n = pgm_read_byte_near(Font);
  Font++;
  while (n > 0) {
    b = pgm_read_byte_near(Font);
    Font++;
    if (sz > 8) {
      b += pgm_read_byte_near(Font) << 8;
      Font++;
    }
    for (j = 0; j <= sz; j++) {
      if (b & 1)
        SetPixA(Cursorx, Cursory + desc - j);
      b = b >> 1;
    }
    Cursorx++;
    n--;
  }
  Cursorx++;
  SendBufxy(px, Cursory + desc - sz, Cursorx - 2, Cursory + desc);
}

void DrawString(char * s, word Font) {
  for (char * t = s; * t; t++) {
    DrawChar( * t, Font);
  }
}

void DrawInt(int i, word Font) {
  if (i < 0) {
    i=-i;
    DrawChar('-',Font);
  }

  bool HasDigit = false;
  int n =  10000;
  if (i == 0) {
    DrawChar('0',Font);
  } else {
    while (n > 0) {
      if ((i > n) or HasDigit) {
        DrawChar('0' + (i / n),Font);
        HasDigit = true;
      }
      i %= n;
      n /= 10;
    }
  }
}

void SetInverted(bool inv) {
  if (inv)
    UC1701SendCommand(0xA6);
  else
    UC1701SendCommand(0xA7);
}

void SetContrast(uint8_t value) // in range 0..63
{
  UC1701SendCommand(0x81);
  UC1701SendCommand(value & 0x3F);
}

void SetResistor(uint8_t value) // in range 0..7
{
  UC1701SendCommand(0x20 + (value & 7));
}

void Initialise() {
  pinMode(pinCS, OUTPUT);
  pinMode(pinCD, OUTPUT);

  digitalWrite(pinCS, 0);

  MakeMark = true;

  digitalWrite(pinCS, 0);
  digitalWrite(pinCD, 0);
  if (pinRST < 255) {
    pinMode(pinRST, OUTPUT);
    digitalWrite(pinRST, 0);
    delay(1);
    digitalWrite(pinRST, 1);
    delay(1);
  }
  digitalWrite(pinCS, 1);
  digitalWrite(pinCS, 0);

  UC1701SendCommand(0xE2);   // System Reset
  delay(10);
  UC1701SendCommand(0xAE);   // Set Display Enable - off
  UC1701SendCommand(0x40);   // Set Scroll Line
  if (curxFlip) {
    UC1701SendCommand(0xA1); // Set SEG Direction - on
    xOffset = 4;
  } else {
    UC1701SendCommand(0xA0); // Set SEG Direction - off
    xOffset = 0;
  }
  if (curyFlip) {
    UC1701SendCommand(0xC8); // Set COM Direction - on
  } else {
    UC1701SendCommand(0xC0); // Set COM Direction - off
  }
  UC1701SendCommand(0xA6);   // Set Inverse Display - off
  UC1701SendCommand(0xA2);   // Set LCD Bias Ratio
  UC1701SendCommand(0x2F);   // Set Power Control - on
  delay(10);
  UC1701SendCommand(0x27);   // Set VLCD Resistor Ratio
  UC1701SendCommand(0x81);   // Set Electronic Volume
  UC1701SendCommand(0x0E);   // value
  UC1701SendCommand(0xFA);   // Set Adv. Program Control
  UC1701SendCommand(0x90);   // value
  UC1701SendCommand(0xAE);   // Set Display Enable - on
  UC1701SendCommand(0xA5);   // Set All-Pixel-ON

  UC1701ClearDisplay();
}

void UC1701Begin(uint8_t CS = 10, uint8_t CD = 7, uint8_t RST = 6, bool xFlip = false, bool yFlip = false) {
  SPI.begin();
  SPI.beginTransaction(SPISettings (4000000UL, MSBFIRST, SPI_MODE0));  // 4 MHz clock, MSB first, mode 0

  pinCS = CS;
  pinCD = CD;
  pinRST = RST;
  curxFlip = xFlip;
  curyFlip = yFlip;
  Initialise();
}

void SetEnabled(bool en) {
  if (en) {
    Initialise();
  } else {
    UC1701SendCommand(0x28);   // Set Power Control - off
    delay(10);
    UC1701SendCommand(0xAE);   // Set Display Enable - off
//    UC1701SendCommand(0xA5);   // All Pixel ON
    if (pinRST < 255)
      digitalWrite(pinRST, 0);
    digitalWrite(pinCS, 0);
    digitalWrite(pinCD, 0);
    delay(1);
  }
}


